/*
  EEPROM.cpp - EEPROM library
  Copyright (c) 2014 Chris Drome.  All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <SPI.h>
#include <Ethernet.h>
#include <SoftwareSerial.h>

/*#define DEBUG*/

#define RFID_MSG_START 0x02
#define RFID_MSG_END 0x03
#define RFID_LENGTH 16

#define RFID_MSG_VALID 0
#define RFID_MSG_INVALID 1
#define RFID_CHECKSUM_PASS 0
#define RFID_CHECKSUM_FAIL 2

#define HEARTBEAT_FREQ 60000UL

/* MAC address of the arduino */
byte MAC_ADDRESS[6] = { 0x90, 0xA2, 0xDA, 0x0D, 0x14, 0xF0 };

/* IP address of the arduino */
const IPAddress IP_ADDRESS (192, 168, 1, 90);

/* IP address of the server */
const IPAddress SERVER_IP_ADDRESS (192, 168, 1, 80);
const int SERVER_PORT = 80;

EthernetClient client;

SoftwareSerial reader(2, 3);
byte buffers[RFID_LENGTH];
int index;

/* Heartbeat */
unsigned int ticks = 0U;
unsigned long last_heartbeat = 0UL;

/* HTTP headers */
char http_heartbeat_header[128];
char http_read_header[160];

/*
Returns RFID_CHECKSUM_PASS if the checksum matches otherwise RFID_CHECKSUM_FAIL.
buffer should point to a valid RFID message.
*/
int is_valid_checksum (byte *buffer)
{
  byte code[6];
  byte checksum = 0;
  byte tempbyte = 0;

  for (int i = 0; i < 12; i++)
  {
    byte val = buffer[i + 1];

    if ((val >= '0') && (val <= '9'))
    {
      val = val - '0';
    }
    else if ((val >= 'A') && (val <= 'F'))
    {
      val = 10 + val - 'A';
    }

    if (i & 1 == 1)
    {
      code[i >> 1] = (val | (tempbyte << 4));

      if (i >> 1 != 5)
      {
        checksum ^= code[i >> 1];
      }
    }
    else
    {
      tempbyte = val;
    }
  }

  return code[5] == checksum ? 0 : -1;
}

/*
Returns RFID_MSG_VALID if buffer points to a valid message otherwise RFID_MSG_INVALID.
*/
int is_valid_rfid_message (byte *buffer)
{
  if (buffer[0] == RFID_MSG_START && buffer[RFID_LENGTH - 1] == RFID_MSG_END)
  {
    for (int i = 1; i < RFID_LENGTH - 1; i++)
    {
      if (buffer[i] == RFID_MSG_START || buffer[i] == RFID_MSG_END)
      {
        return RFID_MSG_INVALID;
      }
    }

    return is_valid_checksum (buffer);
  }

  return RFID_MSG_INVALID;
}

/*
Scans buffer for RFID_MSG_START and shifts the remainder of the buffer to the beginning.
Returns the new index for appending additional bytes.
*/
int shift_rfid_message (byte *buffer)
{
  int head = 0;
  int tail = 0;

  for (int i = 1; i < RFID_LENGTH; i++)
  {
    if (buffer[i] == RFID_MSG_START)
    {
      head = i;
      break;
    }
  }

  if (head > 0)
  {
    for (int i = head; i < RFID_LENGTH; i++)
    {
      buffer[tail] = buffer[i];
      tail++;
    }
  }

  return tail;
}

boolean establish_connection ()
{
  if (! client.connected ())
  {
    /* Attempt a reconnect after closing existing connection */
    client.flush ();
    client.stop ();

    if (! client.connect (SERVER_IP_ADDRESS, SERVER_PORT))
    {
      Serial.print ("Failed to connect to server: ");
      Serial.print (SERVER_IP_ADDRESS);
      Serial.print (":");
      Serial.println (SERVER_PORT);

      return false;
    }
  }

  return client.connected ();
}

void send_heartbeat ()
{
  last_heartbeat = millis ();

  Serial.println ("Sending heartbeat ...");

  if (establish_connection ())
  {
#ifdef DEBUG
    Serial.print (http_heartbeat_header);
#endif /*DEBUG*/

    client.print (http_heartbeat_header);

    while (client.connected () || client.available ())
    {
      if (client.available ())
      {
        char c = client.read ();
#ifdef DEBUG
        Serial.print (c);
#endif /* DEBUG */
      }
    }

    client.flush ();
    client.stop ();
  }
}

void send_read (int head, byte *buffer)
{
  char rfid[11];

  memset (rfid, 0, sizeof (rfid));

  for (int i = 0; i < 10; i++)
  {
    rfid[i] = buffer[i + 1];
  }

  Serial.print ("Sending read ");
  Serial.print (rfid);
  Serial.println (" ...");

  if (establish_connection ())
  {
    char content_length[32];
    char data[32];

    memset (data, 0, sizeof (data));
    snprintf (data, 32, "head=%x&rfid=%s", head, rfid);

    memset (content_length, 0, sizeof (content_length));
    snprintf (content_length, 32, "Content-Length: %d\n\n", strlen (data));

#ifdef DEBUG
    Serial.print (http_read_header);
    Serial.print (content_length);
    Serial.print (data);
#endif /* DEBUG */

    client.print (http_read_header);
    client.print (content_length);
    client.print (data);

    while (client.connected () || client.available ())
    {
      if (client.available ())
      {
        char c = client.read ();
#ifdef DEBUG
        Serial.print (c);
#endif /* DEBUG */
      }
    }

    client.flush ();
    client.stop ();
  }
}

/*
Initialize the sensor.
Set the sensor id, start the software serial, format the message string.
*/
void initialize_sensor ()
{
  Serial.println ("Initializing sensor ...");

  /* Configure the software serialization */
  reader.begin (9600);
  index = 0;

  for (int i = 0; i < RFID_LENGTH; i++)
  {
    buffers[i] = 0;
  }

  Serial.println ("  Starting head: 0");

  delay (500);
}

/*
Initialize the ethernet board.
Read the MAC address from EEPROM, start the ethernet.
*/
void initialize_ethernet ()
{
  char s[18];

  Serial.println ("Initializing ethernet ...");

  memset (s, 0, sizeof (char) * sizeof (s));
  snprintf (s, 18, "%02x:%02x:%02x:%02x:%02x:%02x", MAC_ADDRESS[0], MAC_ADDRESS[1], MAC_ADDRESS[2], MAC_ADDRESS[3], MAC_ADDRESS[4], MAC_ADDRESS[5]);

  Ethernet.begin (MAC_ADDRESS, IP_ADDRESS);

  Serial.print ("  MAC address: ");
  Serial.println (s);
  Serial.print ("  IP address: ");
  Serial.println (Ethernet.localIP ());

  /* Let the ethernet board initialize */
  delay (500);
}

void initialize_http_headers ()
{
  char mac[19];
  char host[22];

  memset (mac, 0, sizeof (mac));
  snprintf (mac, 19, "%02x:%02x:%02x:%02x:%02x:%02x", MAC_ADDRESS[0], MAC_ADDRESS[1], MAC_ADDRESS[2], MAC_ADDRESS[3], MAC_ADDRESS[4], MAC_ADDRESS[5]);

  memset (host, 0, sizeof (host));
  snprintf (host, 22, "%d.%d.%d.%d:%d", SERVER_IP_ADDRESS[0], SERVER_IP_ADDRESS[1], SERVER_IP_ADDRESS[2], SERVER_IP_ADDRESS[3], SERVER_PORT);

  memset (http_heartbeat_header, 0, sizeof (http_heartbeat_header));
  snprintf (http_heartbeat_header, sizeof (http_heartbeat_header), "PUT /rest/v1/system/reader/heartbeat/%s HTTP/1.1\nHost: %s\nConnection: close\n\n", mac, host);

  memset (http_read_header, 0, sizeof (http_read_header));
  snprintf (http_read_header, sizeof (http_read_header), "POST /rest/v1/system/reader/read/%s HTTP/1.1\nHost: %s\nConnection: close\nContent-Type: text/plain\n", mac, host);
}

/*
Initialize the connection to the server.
*/
void initialize_connection ()
{
  char host[24];

  Serial.println ("Initializing server connection ...");

  memset (host, 0, sizeof (host));
  snprintf (host, sizeof (host), "%d.%d.%d.%d:%d", SERVER_IP_ADDRESS[0], SERVER_IP_ADDRESS[1], SERVER_IP_ADDRESS[2], SERVER_IP_ADDRESS[3], SERVER_PORT);

  if (client.connect (SERVER_IP_ADDRESS, SERVER_PORT))
  {
    Serial.print ("  Connection established: ");
    Serial.println (host);

    send_heartbeat ();
  }
  else
  {
    Serial.print ("  Connection failed: ");
    Serial.println (host);
  }
}

void setup ()
{
  Serial.begin (9600);

  initialize_sensor ();
  initialize_ethernet ();
  initialize_http_headers ();
  initialize_connection ();

  Serial.println ("Initialization complete");
}

void loop ()
{
  while (reader.available () > 0)
  {
    int c = reader.read ();

    if (c >= 0 && c < 255)
    {
      /* Reset index to beginning of buffer when message start is found. */
      if (c == RFID_MSG_START)
      {
        index = 0;
      }

      buffers[index] = c;
      index++;
    }

    if (index == RFID_LENGTH)
    {
      if (is_valid_rfid_message (buffers) == RFID_MSG_VALID)
      {
        send_read (0, buffers);
        index = 0;
      }
      else
      {
        index = shift_rfid_message (buffers);
        Serial.println ("Invalid RFID message");
      }
    }
  }

  ticks++;

  /* Check millis roughly once every couple seconds */
  if (ticks >= 40000U)
  {
    ticks = 0U;
    unsigned long ts = millis ();

    if (ts > last_heartbeat)
    {
      if (ts - last_heartbeat > HEARTBEAT_FREQ)
      {
        send_heartbeat ();
      }
    }
    else
    {
      if (~(last_heartbeat - ts) > HEARTBEAT_FREQ)
      {
        send_heartbeat ();
      }
    }
  }
}
